/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2010-2014 Intel Corporation
 */

#ifndef __INCLUDE_RTE_LRU_LOONGARCH_H__
#define __INCLUDE_RTE_LRU_LOONGARCH_H__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

#include <rte_config.h>
#include <rte_vect.h>

#ifndef RTE_TABLE_HASH_LRU_STRATEGY
#define RTE_TABLE_HASH_LRU_STRATEGY                        3
#endif

#if RTE_TABLE_HASH_LRU_STRATEGY == 3

#define lru_init(bucket)						\
	{ bucket->lru_list = ~0LLU; }

static inline int
f_lru_pos(uint64_t lru_list)
{
	xmm_t lst = (xmm_t) {.i64 = { lru_list, (uint64_t)-1} };
	uint16_t min_val, i, j = 0;

	min_val = lst.u16[0];
	for (i = 0; i < 8; i++) {
		if (min_val > lst.u16[i]) {
			min_val = lst.u16[i];
			j = i;
		}
	}

	return j;
}
#define lru_pos(bucket) f_lru_pos(bucket->lru_list)

#define lru_update(bucket, mru_val)					\
do {									\
	const uint64_t orvals[] = {0xFFFFLLU, 0xFFFFLLU << 16,		\
		0xFFFFLLU << 32, 0xFFFFLLU << 48, 0LLU};		\
	const uint64_t decs[] = {0x1000100010001LLU, 0};		\
	__m128i v_zeros = __lsx_vldi(0);				\
	__m128i lru = v_zeros, vdec = v_zeros;				\
	lru = __lsx_vinsgr2vr_d(lru, bucket->lru_list, 0);		\
	vdec = __lsx_vinsgr2vr_d(vdec, decs[mru_val>>2], 0);		\
	lru = __lsx_vsub_h(lru, vdec);					\
	bucket->lru_list = __lsx_vpickve2gr_d(lru, 0);			\
	bucket->lru_list |= orvals[mru_val];				\
} while (0)
#endif

#ifdef __cplusplus
}
#endif

#endif
